///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "SocketServer.h"
#include "MessageAccumulatingSocketServer.h"
#include "FragmentEchoingSocketServer.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/IIPAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/ServerCollection.h"
#include "JetByteTools/SocketTools/FullAddress.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/SequencedStreamSocketAllocator.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/CommandLine.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::IIPAddress;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::CServerCollection;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char *argv[ ])
{
   CSimpleServerShutdownHandler shutdownHandler;

   CGlobalErrorHandler errorHandler;

   int result = 0;

   try
   {
      CAsyncFileLog log(_T("BinaryWebSocketServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CCommandLine commandLine(argc, argv, _T("BinaryWebSocketServer"), CCommandLine::TCPServer | CCommandLine::DoesNotSupportWelcomeMessage);

         if (commandLine.Parse())
         {
            CIOPool pool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            pool.Start();

            CSequencedStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket(),
               commandLine.DisplayDebug());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const ListenBacklog listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            const CAddressType addressPreference = CAddressTypeIPv4::instance;

            IIPAddress::Port port = commandLine.Port();

            CServerCollection servers;

            // HyBi server
            {
               OutputEx(_T("HyBi server on port: ") + ToString(port));

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CSocketServer(
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               OutputEx(_T("HyBi message accumulating server on port: ") + ToString(port));

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CMessageAccumulatingSocketServer(
                  true,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               OutputEx(_T("HyBi message accumulating server on port: ") + ToString(port) + _T(" (no message acccumulation in WebSocket layer)"));

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CMessageAccumulatingSocketServer(
                  false,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               OutputEx(_T("HyBi fragment echoing server on port: ") + ToString(port));

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CFragmentEchoingSocketServer(
                  true,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            {
               OutputEx(_T("HyBi fragment echoing server on port: ") + ToString(port) + _T(" (no message acccumulation in WebSocket layer)"));

               const CFullAddress address(
                  commandLine.Server(),
                  port++,
                  addressPreference);

               servers.AddServer(new CFragmentEchoingSocketServer(
                  false,
                  address,
                  listenBacklog,
                  pool,
                  socketAllocator,
                  bufferAllocator,
                  connectionLimiter));
            }

            servers.Start();

            servers.StartAcceptingConnections();

            shutdownHandler.SetServerControl(servers, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            servers.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
