///////////////////////////////////////////////////////////////////////////////
// File: Connection.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "Connection.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

#include "JetByteTools/IOTools/SmartBuffer.h"

#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/ICreateStreamSocketConnections.h"
#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#include "JetByteTools/WebSocketTools/HyBi/IWebSocket.h"
#include "JetByteTools/WebSocketTools/HyBi/CloseStatusCodes.h"

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::ToString;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::ICreateStreamSocketConnections;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::CAddressRenderer;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CSmartStreamSocket;
using JetByteTools::Socket::ShutdownSend;
using JetByteTools::Socket::ShutdownReceive;

using JetByteTools::WebSocket::HyBi::IWebSocket;
using JetByteTools::WebSocket::HyBi::CloseStatusNormal;
using JetByteTools::WebSocket::HyBi::CSmartWebSocket;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CConnection
///////////////////////////////////////////////////////////////////////////////

CConnection::CConnection(
   ICreateStreamSocketConnections &connectionCreator,
   IWebSocket &socket,
   const IAddress & /*address*/,
   const IAddress &remoteAddress)
   :  m_connectionCreator(connectionCreator),
      m_pLocalSocket(&socket),
      m_pRemoteSocket(0),
      m_remoteAddress(remoteAddress),
      m_ref(1)
{
   m_pLocalSocket->AddRef();
}

CConnection::~CConnection()
{
   if (m_pLocalSocket)
   {
      m_pLocalSocket->Release();
   }

   if (m_pRemoteSocket)
   {
      m_pRemoteSocket->Release();
   }
}

void CConnection::AddRef()
{
   m_ref.Increment();
}

void CConnection::Release()
{
   const unsigned long ref = m_ref.Decrement();

   if (ref == 0)
   {
      delete this;
   }
}

CSmartWebSocket CConnection::GetLocalSocket()
{
   ILockableObject::Owner lock(m_lock);

   return CSmartWebSocket(m_pLocalSocket, CSmartWebSocket::CreateNewReference);
}

CSmartWebSocket CConnection::DetachLocalSocket()
{
   ILockableObject::Owner lock(m_lock);

   auto pSocket = m_pLocalSocket;

   m_pLocalSocket = 0;

   return CSmartWebSocket(pSocket);
}

CSmartStreamSocket CConnection::GetRemoteSocket()
{
   ILockableObject::Owner lock(m_lock);

   return CSmartStreamSocket(m_pRemoteSocket, CSmartStreamSocket::CreateNewReference);
}

CSmartStreamSocket CConnection::DetachRemoteSocket()
{
   ILockableObject::Owner lock(m_lock);

   auto pSocket = m_pRemoteSocket;

   m_pRemoteSocket = 0;

   return CSmartStreamSocket(pSocket);
}

void CConnection::OnConnectionEstablished()
{
   // The remote connection gets to hold a reference to this object

   AddRef();

   // connect to the remote connection

   m_connectionCreator.ConnectAsyncIfAvailable(m_remoteAddress, this);
}

void CConnection::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_pRemoteSocket = &socket;

      m_pRemoteSocket->AddRef();
   }

   // and now we can issue a read for our local socket

   auto localSocket(GetLocalSocket());

   if (localSocket.Get())
   {
      localSocket->TryRead();
   }

   socket.TryRead();
}

void CConnection::OnConnectionFailed(
   const DWORD lastError)
{
   auto localSocket(DetachLocalSocket());

   if (localSocket.Get())
   {
      const auto message = _T("Connection Failed: ") + ToString(lastError);

      localSocket->Close(4000, message);
   }
}

void CConnection::OnReadCompleted(
   IWebSocket &socket,
   CSmartBuffer &buffer)
{
   auto localSocket(GetLocalSocket());

   if (localSocket.Get())
   {
      if (&socket != localSocket.Get())
      {
         throw CException(_T("CConnection::OnReadCompleted()"), _T("Unexpected socket."));
      }
   }

   {
      auto remoteSocket(GetRemoteSocket());

      if (remoteSocket.Get())
      {
         remoteSocket->TryWrite(buffer);
      }
   }

   if (localSocket.Get())
   {
      localSocket->TryRead();
   }
}

void CConnection::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   auto remoteSocket(GetRemoteSocket());

   if (remoteSocket.Get())
   {
      if (&socket != remoteSocket.Get())
      {
         throw CException(_T("CConnection::OnReadCompleted()"), _T("Unexpected socket."));
      }
   }

   auto localSocket(GetLocalSocket());

   if (localSocket.Get())
   {
      localSocket->TryWriteBinary(buffer);
   }

   if (remoteSocket.Get())
   {
      remoteSocket->TryRead();
   }
}

void CConnection::OnConnectionClosed(
   IWebSocket &socket)
{
   auto localSocket(DetachLocalSocket());

   if (localSocket.Get())
   {
      if (&socket != localSocket.Get())
      {
         throw CException(_T("CConnection::OnConnectionClosed()"), _T("Unexpected socket."));
      }
   }

   auto remoteSocket(DetachRemoteSocket());

   if (remoteSocket.Get())
   {
      remoteSocket->Shutdown(ShutdownSend);
   }

   if (localSocket.Get())
   {
      localSocket->Close(CloseStatusNormal, _T("Client closed connection"));
   }
}

void CConnection::OnConnectionClosed(
   IStreamSocket &socket)
{
   auto remoteSocket(DetachRemoteSocket());

   auto localSocket(DetachLocalSocket());

   if (remoteSocket.Get() && &socket == remoteSocket.Get())
   {
      //JetByteTools::Core::OutputEx(_T("CConnection::OnConnectionClosed - S"));

      remoteSocket->Shutdown(ShutdownReceive);

      auto localSocket(DetachLocalSocket());

      if (localSocket.Get())
      {
         localSocket->Close(CloseStatusNormal, _T("Remote connection closed"));
      }
   }
   else if (localSocket.Get())
   {
      //JetByteTools::Core::OutputEx(_T("CConnection::OnConnectionClosed - WS"));

      if (remoteSocket.Get())
      {
         remoteSocket->Shutdown(ShutdownReceive);
      }
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
