///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"
#include "Connection.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#include "JetByteTools/WebSocketTools/AutoDetectProtocolHandler.h"
#include "JetByteTools/WebSocketTools/Headers.h"
#include "JetByteTools/WebSocketTools/HyBi/ProtocolHandler.h"
#include "JetByteTools/WebSocketTools/HyBi/CloseStatusCodes.h"
#include "JetByteTools/WebSocketTools/HyBi/RequestHeaders.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::ErrorCodeToErrorMessage;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ConnectionClosureReason;

using JetByteTools::WebSocket::CHeaders;
using JetByteTools::WebSocket::MessageStatus;
using JetByteTools::WebSocket::MessageType;
using JetByteTools::WebSocket::ConnectionEstablishmentResult;
using JetByteTools::WebSocket::ConnectionEstablished;
using JetByteTools::WebSocket::IProtocolHandler;

using JetByteTools::WebSocket::HyBi::CProtocolHandler;
using JetByteTools::WebSocket::HyBi::IWebSocket;
using JetByteTools::WebSocket::HyBi::CloseStatusNormal;
using JetByteTools::WebSocket::HyBi::CRequestHeaders;

using std::string;
using std::wstring;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   const IFullAddress &remoteAddress,
   const IFullAddress &localAddress,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(localAddress, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_remoteAddress(remoteAddress),
      m_protocolHandlerIndex(socketAllocator.RequestUserDataSlot(_T("m_protocolHandlerIndex"))),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CSocketServer - m_userDataIndex"))),
      m_bufferAllocator(bufferAllocator),
      m_protocolHandlerAllocator(bufferAllocator)
{

}

CSocketServer::~CSocketServer()
{
   WaitForShutdownToComplete();
}

void CSocketServer::OnPreOutgoingConnect(
   IIndexedOpaqueUserData &userData,
   const IAddress & /*address*/,
   const void *pUserData)
{
   userData.SetUserPointer(m_userDataIndex, const_cast<void *>(pUserData));
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   Output(_T("OnConnectionEstablished"));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   if (pConnection)
   {
      pConnection->OnConnectionEstablished(socket, address);
   }
   else
   {
      const DWORD flags = CProtocolHandler::InsecureConnection | CProtocolHandler::ServerMaskingRequirements;

      auto pHandler = m_protocolHandlerAllocator.Allocate(*this, socket, flags);

      socket.SetUserPointer(m_protocolHandlerIndex, pHandler);

      pHandler->AcceptHandshake();
   }
}

void CSocketServer::OnOutgoingConnectionFailed(
   IStreamSocket &socket,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   Output(_T("OnOutgoingConnectionFailed: ") + ErrorCodeToErrorMessage(lastError));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   if (pConnection)
   {
      pConnection->OnConnectionFailed(lastError);
   }
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      auto pHandler = reinterpret_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

      if (pHandler)
      {
         DEBUG_ONLY(Output(_T("WebSocket Data: \r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

         pHandler->OnDataReceived(buffer);
      }
      else
      {
         DEBUG_ONLY(Output(_T("TCP Data: \r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

         auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

         pConnection->OnReadCompleted(socket, buffer);
      }
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CSocketServer::OnConnectionClientClose(
   IStreamSocket &socket)
{
   Output(_T("OnConnectionClientClose"));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   pConnection->OnConnectionClosed(socket);

   auto pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   if (pHandler)
   {
      pHandler->OnConnectionClosed();
   }
}

void CSocketServer::OnConnectionReset(
   IStreamSocket &socket,
   const DWORD /*lastError*/)
{
   Output(_T("OnConnectionReset"));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   pConnection->OnConnectionClosed(socket);

   auto pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   if (pHandler)
   {
      pHandler->OnConnectionClosed();
   }
}

void CSocketServer::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason /*reason*/)
{
   OutputEx(_T("OnConnectionClosure"));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   pConnection->OnConnectionClosed(socket);

   auto pHandler = reinterpret_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   if (pHandler)
   {
      pHandler->OnConnectionClosed();
   }
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   OutputEx(_T("OnSocketReleased"));

   auto pHandler = reinterpret_cast<IProtocolHandler *>(userData.GetUserPointer(m_protocolHandlerIndex));

   delete pHandler;

   auto pConnection = reinterpret_cast<CConnection *>(userData.GetUserPointer(m_userDataIndex));

   pConnection->Release();
}

void CSocketServer::OnError(
   const _tstring &message)
{
   OutputEx(_T("OnError: \"") + message + _T("\""));
}

void CSocketServer::OnOutboundConnectionEstablished(
   IWebSocket & /*socket*/,
   const CHeaders & /*responseHeaders*/)
{
   throw CException(_T("CSocketServer::OnOutboundConnectionEstablished()"), _T("Unexpected outbound websocket connection established"));
}

ConnectionEstablishmentResult CSocketServer::OnConnectionHandshake(
   IIndexedOpaqueUserData & /*userData*/,
   const string &uri,
   const bool /*secure*/,
   const CHeaders &requestHeaders,
   CHeaders &responseHeaders)
{
   Output(_T("OnConnectionHandshake - : ") + CStringConverter::AtoT(uri));

   CRequestHeaders::DumpRequestHeadersToDebug(requestHeaders);

   const string versionHeader = "sec-websocket-version";

   if (requestHeaders.Contains(versionHeader))
   {
      const auto version = requestHeaders.GetValue(versionHeader);

      Output(_T("Protocol Version: HyBi: ") + CStringConverter::AtoT(version));
   }
   else
   {
      throw CException(_T("CHyBiSocketServer::OnConnectionHandshake()"), _T("Protocol version unknown"));
   }

   CRequestHeaders::AddOriginHeaders(responseHeaders, requestHeaders);

   return ::ConnectionEstablished;
}

void CSocketServer::OnConnectionEstablished(
   IWebSocket &socket,
   const string &uri,
   const bool /*secure*/)
{
   Output(_T("OnConnectionEstablished: ") + CStringConverter::AtoT(uri));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   if (!pConnection)
   {
      pConnection = new CConnection(*this, socket, socket.GetRemoteAddress(), m_remoteAddress);

      socket.SetUserPointer(m_userDataIndex, pConnection);
   }

   pConnection->OnConnectionEstablished();
}

void CSocketServer::OnData(
   IWebSocket & /*socket*/,
   const _tstring & /*text*/,
   const MessageStatus /*status*/,
   const __int64 /*messageBytesOutstanding*/)
{
   // For this to be called you need to specify CProtocolHandler::DispatchTextAsStrings in the flags that are passed
   // during construction

   throw CException(
      _T("CSocketServer::OnData()"),
      _T("Unexpected: we don't handle text frames"));
}

void CSocketServer::OnData(
   IWebSocket &socket,
   CSmartBuffer &buffer,
   const MessageType /*type*/,
   const MessageStatus /*status*/,
   const __int64 /*messageBytesOutstanding*/)
{
   DEBUG_ONLY(Output(_T("OnDataFrame - \r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   pConnection->OnReadCompleted(socket, buffer);
}

void CSocketServer::OnPingResponse(
   IWebSocket & /*socket*/,
   const BYTE * /*pData*/,
   const BYTE /*length*/)
{
   throw CException(
      _T("CSocketServer::OnPingResponse()"),
      _T("Unexpected: we don't send pings"));
}

void CSocketServer::OnClientClose(
   IWebSocket &socket,
   const WORD status,
   const _tstring &reason)
{
   OutputEx(_T("OnClientClose: ") + ToString(status) + _T(": ") + reason);

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   pConnection->OnConnectionClosed(socket);

   socket.Close(CloseStatusNormal);
}

void CSocketServer::OnError(
   IWebSocket & /*socket*/,
   const _tstring &message)
{
   OutputEx(_T("OnError:") + message);
}

void CSocketServer::OnClosed(
   IWebSocket & /*socket*/,
   const WORD status)
{
   OutputEx(_T("OnClosed:") + ToString(status));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
