///////////////////////////////////////////////////////////////////////////////
// File: TestMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "JetByteTools/SocketTools/WinsockWrapper.h"

#include "StreamSocketConnectionManager.h"
#include "TestMessages.h"
#include "TestMessage.h"
#include "MessageCollection.h"
#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/PerformanceCounter.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/Win32Tools/GlobalName.h"
#include "JetByteTools/Win32Tools/ManualResetEvent.h"
#include "JetByteTools/Win32Tools/SecurityDescriptorAllowAll.h"
#include "JetByteTools/Win32Tools/SecurityAttributes.h"

#include "JetByteTools/IOTools/IOPool.h"
#include "JetByteTools/IOTools/BufferAllocator.h"
#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"
#include "JetByteTools/SocketTools/SequencedStreamSocketAllocator.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#include "JetByteTools/OpenSSLTools/Context.h"
#include "JetByteTools/OpenSSLTools/Exception.h"

#pragma hdrstop

#include <openssl/rand.h>
#include <openssl/err.h>

#include <iostream>

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CException;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CPerformanceCounter;
using JetByteTools::Core::CGlobalErrorHandler;
using JetByteTools::Core::GetStringLength;

using JetByteTools::Win32::CManualResetEvent;
using JetByteTools::Win32::CGlobalName;
using JetByteTools::Win32::CSecurityDescriptorAllowAll;
using JetByteTools::Win32::CSecurityAttributes;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CUsesWinsock;
using JetByteTools::Socket::CSmartStreamSocket;
using JetByteTools::Socket::CSequencedStreamSocketAllocator;

using JetByteTools::IO::CIOPool;
using JetByteTools::IO::CBufferAllocator;
using JetByteTools::IO::CAsyncFileLog;
using JetByteTools::IO::CPendingWriteLimiter;

using JetByteTools::OpenSSL::CContext;

using std::cout;
using std::cin;
using std::endl;
using std::min;
using std::string;

///////////////////////////////////////////////////////////////////////////////
// File level static variables
///////////////////////////////////////////////////////////////////////////////

// Note that this can be set from the command line using the -password command
// line argument. This should be the password that was used when you created
// your server's key file - see notes at the end of this file for how to
// create a self signed certificate and key.

static string clientKeyPassword = "password";

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static SSL_CTX *CreateSSLContext(
   bool useClientCertificates,
   const _tstring &certificatePath,
   bool verifyPeer);

static void BuildMessagesForConnection(
   CTestMessage::DataLength messageSize,
   CTestMessages &messages,
   bool textMessages);

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char **argv)
{
   auto result = 2;

   CGlobalErrorHandler errorHandler;

   try
   {
      // Note that maxPendingWrites can be used to limit the number of pending
      // write operations to prevent uncontrolled 'non paged pool' memory usage 
      // if log messages are produced at a rate faster than they can be written 
      // to the disk. Note that once the limit is reached all writes become 
      // synchronous until the number of pending writes has dropped below the 
      // limit again. This is only ever likely to be an issue in a debug build
      // but it WILL affect performance and thread scheduling.

      static const long maxPendingWrites = 100000;

      CPendingWriteLimiter limiter(maxPendingWrites);

      CAsyncFileLog log(_T("WebSocketEchoServerTest.log"), limiter);

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CUsesWinsock usesWinsock;

         CCommandLine commandLine;

         if (commandLine.Parse(argc, argv))
         {
            result = 1;

            const auto messageRepeatCount = commandLine.Messages();

            const auto numConnections = commandLine.Connections();

            const auto transferRate = commandLine.MessageRate();

            const CBufferAllocator::BufferSize messageSize = commandLine.MessageSize();

            const auto delayAfter = commandLine.ConnectionBatchSize();

            const auto delayFor = commandLine.ConnectionBatchPause();

            const auto sendAfterRecv = commandLine.SendAfterRecv();

            const auto syncConnect = commandLine.SyncConnect();

            const auto holdConnections = commandLine.HoldConnections();

            const auto maxSocketsInPool = numConnections;

            const CBufferAllocator::BufferCount maxBuffersInPool = numConnections * 2;

            const CBufferAllocator::BufferSize bufferSize = commandLine.MessageSize() * 2;

            const auto fragments = commandLine.Fragments();

            const auto multipleMessages = commandLine.MultipleMessages();

            const auto dataFlowTimeoutMillis = commandLine.MessageFlowTimeout();

            const auto displayConnectionBatching = commandLine.DisplayConnectionBatching();

            clientKeyPassword = CStringConverter::TtoA(commandLine.Password());

            const auto sslEnabled = commandLine.UseSSL();

            CContext sslContext(
               CreateSSLContext(
                  commandLine.UseClientCerts(),
                  commandLine.CertificatePath(),
                  commandLine.VerifyServerCert()));

           CSequencedStreamSocketAllocator socketAllocator(
               maxSocketsInPool,
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket());

            CBufferAllocator bufferAllocator(
               bufferSize,
               maxBuffersInPool);

            CIOPool pool(
               commandLine.NumberOfIOThreads());

            pool.Start();

            CStreamSocketConnectionManager connector(
               sslEnabled,
               sslContext,
               commandLine.VerifyServerCert(),
               commandLine.Hixie(),
               commandLine.ReadTimeout(),
               pool,
               socketAllocator,
               bufferAllocator);

            const auto addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            if (!connector.AsyncConnectAvailable(address.Type()) && !syncConnect)
            {
               commandLine.ShowHelp(
                  _T("ConnectEx() is not available on this platform.\n")
                  _T("Use -syncConnect to force synchronous connections"));

               return 1;
            }

            if (commandLine.PreAllocate())
            {
               socketAllocator.EnlargePool(maxSocketsInPool);

               bufferAllocator.EnlargePool(maxBuffersInPool);
            }

            // now loop opening connections and kicking things off...

            CMessageCollection messages(numConnections);

            try
            {
               // build messages here and pass in to each connection

               if (messageRepeatCount != 0)
               {
                  cout << "Building Messages" << endl;
               }

               size_t i = 0;

               const CBufferAllocator::BufferSize skipSignOnMessageBytes = commandLine.SkipSignOnMessageBytes();

               for (i = 0; i < numConnections; ++i)
               {
                  auto pMessages = new CTestMessages(
                     i,
                     messageRepeatCount,
                     transferRate,
                     messages,
                     sendAfterRecv,
                     holdConnections,
                     skipSignOnMessageBytes,
                     fragments,
                     multipleMessages,
                     CTestMessages::TrackPendingMessages,
                     commandLine.TextMessages());

                  BuildMessagesForConnection(messageSize, *pMessages, commandLine.TextMessages());

                  messages.Add(pMessages);
               }

               cout << "Creating " << numConnections << " connections" << endl;

               i = 0;

               CPerformanceCounter counter(CPerformanceCounter::LockThreadToOneCPU);

               auto ok = true;

               for (auto it = messages.Begin(); ok && it != messages.End(); ++it)
               {
                  auto pMessages = *it;

                  if (syncConnect)
                  {
                     if (sslEnabled)
                     {
                        auto socket = connector.SecureConnectNoThrow(address, pMessages);

                        ok = (socket.Get() != nullptr);
                     }
                     else
                     {
                        auto socket = connector.ConnectNoThrow(address, pMessages);

                        ok = (socket.Get() != nullptr);
                     }
                  }
                  else
                  {
                     if (sslEnabled)
                     {
                        auto socket = connector.SecureAsyncConnectNoThrow(address, pMessages);

                        ok = (socket.Get() != nullptr);
                     }
                     else
                     {
                        auto socket = connector.AsyncConnectNoThrow(address, pMessages);

                        ok = (socket.Get() != nullptr);
                      }
                  }

                  if (delayAfter && (++i % delayAfter == 0))
                  {
                     if (displayConnectionBatching)
                     {
                        cout << i << " connections created" << endl;
                     }

                     if (delayFor)
                     {
                        Sleep(delayFor);
                     }
                  }
               }

               if (ok)
               {
                  cout << "All connections in progress" << endl;
               }
               else
               {
                  SetEvent(messages.GetConnectionsCompleteEvent());
               }

               CSecurityDescriptorAllowAll sd;

               CSecurityAttributes sa(sd);

               CManualResetEvent shutdownEvent(CGlobalName(_T("JetByteToolsServerShutdown")), &sa);

               HANDLE handlesToWaitFor[2];

               handlesToWaitFor[0] = shutdownEvent.GetWaitHandle();
               handlesToWaitFor[1] = messages.GetConnectionsCompleteEvent();

               auto waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, INFINITE);

               if (waitResult == WAIT_OBJECT_0)
               {
                  result = 1;
               }
               else if (waitResult == WAIT_OBJECT_0 + 1)
               {
                  cout << "All connections complete in " << CStringConverter::TtoA(counter.GetElapsedTimeAsString()) << endl;
                  cout << messages.ConnectionsEstablished() << " established. " << messages.ConnectionsFailed() << " failed."<< endl;

                  if (messages.ConnectionsEstablished())
                  {
                     if (messages.ConnectionsFailed() != 0)
                     {
                        result = 1;
                     }

                     if (messageRepeatCount != 0)
                     {
                        cout << "Waiting for data flow to complete" << endl;

                        const auto checkEvery = commandLine.CheckEvery();

                        Milliseconds elapsedTime = 0;

                        auto done = false;

                        while (!done)
                        {
                           handlesToWaitFor[1] = messages.GetMessagesCompleteEvent();

                           const auto timeout = (dataFlowTimeoutMillis == INFINITE || checkEvery == INFINITE ? INFINITE : min(checkEvery, dataFlowTimeoutMillis - elapsedTime));

                           elapsedTime += timeout;

                           const auto waitResult = ::WaitForMultipleObjects(2, handlesToWaitFor, false, timeout);

                           if (waitResult == WAIT_TIMEOUT)
                           {
                              if (elapsedTime >= dataFlowTimeoutMillis)
                              {
                                 cout << "Waiting for data flow to complete timed out after " << CStringConverter::TtoA(counter.GetElapsedTimeAsString()) << endl;

                                 result = 1;
                                 done = true;
                              }
                              else
                              {
                                 cout << "Waiting for " << messages.GetNumberOfConnectionsWithMessagesPending() << " connections to complete" << endl;

                                 CMessageCollection::StringErrors errors;

                                 if (!messages.CheckMessages(errors))
                                 {
                                    if (!errors.empty())
                                    {
                                       cout << "Message errors:" << endl;

                                       for (CMessageCollection::StringErrors::const_iterator it = errors.begin(); it != errors.end(); ++it)
                                       {
                                          cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
                                       }
                                    }
                                 }
                              }
                           }
                           else if (waitResult == WAIT_OBJECT_0)
                           {
                              result = 1;
                              done = true;
                           }
                           else if (waitResult == WAIT_OBJECT_0 + 1)
                           {
                              const auto elapsedTime = counter.GetElapsedTimeAsString();

                              cout << "Data flow complete in " << CStringConverter::TtoA(elapsedTime) << endl;

                              cout << "Average RTT: " << messages.GetAverageRTT() << "ms" << endl;
                              cout << "    Max RTT: " << messages.GetMaxRTT() << "ms" << endl;
                              cout << "    Min RTT: " << messages.GetMinRTT() << "ms" << endl;

                              result = 0;
                              done = true;
                           }
                           else
                           {
                              throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
                           }
                        }
                     }
                     else
                     {
                        result = 0;
                     }
                  }
               }
               else
               {
                  throw CException(_T("main()"), _T("Unexpected result from WaitForMultipleObjects"));
               }

               if (commandLine.HoldConnections() && commandLine.IsInteractive() && messages.ConnectionsEstablished())
               {
                  cout << "Press return to close connections" << endl;

                  char input;

                  cin >> std::noskipws >> input;
               }
            }
            catch(const CException &e)
            {
               cout << "Exception: " << e.GetDetailsA() << endl;

               result = 1;
            }
            catch(...)
            {
               cout << "Unexpected exception" << endl;

               result = 1;
            }

            if (messageRepeatCount != 0)
            {
               CMessageCollection::StringErrors errors;

               if (!messages.CheckMessages(errors))
               {
                  result = 1;
               }

               if (!errors.empty())
               {
                  cout << "Message errors:" << endl;

                  for (CMessageCollection::StringErrors::const_iterator it = errors.begin(); it != errors.end(); ++it)
                  {
                     cout << ToStringA(it->second) << " - " << CStringConverter::TtoA(it->first) << endl;
                  }

                  result = 1;
               }
            }

            cout << "Wait for connector shutdown..." << endl;

            if (!connector.WaitForShutdownToComplete(10000))
            {
               cout << "Connector shutdown timed out!" << endl;

               result = 1;

               connector.ForceShutdown();
            }

            cout << "Wait for IO pool shutdown..." << endl;

            pool.WaitForShutdownToComplete();

            cout << "Waiting for socket allocator to flush..." << endl;

            socketAllocator.ReleaseSockets();

            cout << "Waiting for buffer allocator to flush..." << endl;

            bufferAllocator.Flush();

            if (!connector.DumpErrors())
            {
               result = 1;
            }

            cout << "All done..." << endl;
         }
      }
      catch(...)
      {
         // TODO
      }
   }
   catch(const CException &e)
   {
      cout << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(const char *pMsg)
   {
      cout << pMsg << endl;

      result = 1;
   }
   catch(...)
   {
      cout << "Unexpected exception" << endl;

      result = 1;
   }

   cout << "Test " << (result == 0 ? "Passed" : "Failed") << endl;

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static void BuildMessagesForConnection(
   const CTestMessage::DataLength messageSize,
   CTestMessages &messages,
   const bool textMessages)
{
   messages.AddMessage(new CTestMessage(messageSize, textMessages));
}

///////////////////////////////////////////////////////////////////////////////
// OpenSSL specific stuff.
//
// Note: The following code just sets up an example SSL Context, it's been
// put together from code samples in the OpenSSL distribution. You should read
// and understand the OpenSSL documentation and then write some real code to
// initialise your context in the way that you require.
///////////////////////////////////////////////////////////////////////////////

// This callback function is used to return the password for the server's key.
// In this example we simply return the value that was passed on the command line
// with the -password command line arg, or "password" if none is specified.

static int password_cb(
   char *buf,
   int num,
   int /*rwflag*/,
   void * /*userdata*/)
{
   const auto length = GetStringLength<int>(clientKeyPassword);

   if (num < (length + 1))
   {
      return 0;
   }

   strcpy_s(buf, num, clientKeyPassword.c_str());

   return length;
}

static SSL_CTX *CreateSSLContext(
   const bool useClientCertificates,
   const _tstring &certificatePath,
   const bool verifyPeer)
{
   if (1 != RAND_status())
   {
      // this is now done automatically, see OpenSSL documentation if you ever get this message!

      throw JetByteTools::OpenSSL::CException(
         _T("CreateSSLContext() - random number generator not seeded"),
         0);
   }

   const SSL_METHOD *pMethod = TLS_client_method();

   SSL_CTX *pContext = SSL_CTX_new(pMethod);

   const string cipherList("ALL");

   if (0 == SSL_CTX_set_cipher_list(pContext, cipherList.c_str()))
   {
      const auto error = ERR_get_error();

      throw JetByteTools::OpenSSL::CException(
         _T("CreateSSLContext() - Failed to set cipher list to \"") +
         CStringConverter::AtoT(cipherList) + _T("\""), error);
   }

   if (useClientCertificates)
   {
      const auto clientPemFile = CStringConverter::TtoA(certificatePath) + "client.pem";

      if (!(SSL_CTX_use_certificate_chain_file(pContext, clientPemFile.c_str())))
      {
         const auto error = ERR_get_error();

         throw JetByteTools::OpenSSL::CException(
            _T("CreateSSLContext() - Failed to load client certificate file: \"") +
            CStringConverter::AtoT(clientPemFile) + _T("\""), error);
      }

      SSL_CTX_set_default_passwd_cb(pContext, password_cb);

      if (!(SSL_CTX_use_PrivateKey_file(pContext, clientPemFile.c_str(), SSL_FILETYPE_PEM)))
      {
         const auto error = ERR_get_error();

         // Most common cause of failure here is that the password doesn't match...

         throw JetByteTools::OpenSSL::CException(
            _T("CreateSSLContext() - Failed to load private key file: \"") +
            CStringConverter::AtoT(clientPemFile) + _T("\""), error);
      }
   }

   if (verifyPeer)
   {
      // To be able to verify a peer certificate we need a set of trusted root certificates...

      const auto rootPemFile = CStringConverter::TtoA(certificatePath) + "root.pem";

      /* Load the CAs we trust*/
      if(!(SSL_CTX_load_verify_locations(pContext, rootPemFile.c_str(), nullptr)))
      {
         const auto error = ERR_get_error();

         throw JetByteTools::OpenSSL::CException(
            _T("CreateSSLContext() - Failed to load root CA file: \"") +
            CStringConverter::AtoT(rootPemFile) + _T("\""), error);
      }

      if (!SSL_CTX_set_default_verify_paths(pContext))
      {
         const auto error = ERR_get_error();

         throw JetByteTools::OpenSSL::CException(
            _T("CreateSSLContext() - Failed to set verification paths"), error);
      }
   }

   return pContext;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
