///////////////////////////////////////////////////////////////////////////////
// File: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "SocketServer.h"
#include "CommandLine.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ZeroInitialiseExpandableBuffer.h"
#include "JetByteTools/CoreTools/GlobalErrorHandler.h"

#include "JetByteTools/IOTools/AsyncFileLog.h"

#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/AddressTypeIPv4.h"

#include "JetByteTools/OpenSSLTools/Context.h"
#include "JetByteTools/OpenSSLTools/Exception.h"

#include "ServerCommon/IOPool.h"
#include "ServerCommon/SequencedStreamSocketAllocator.h"
#include "ServerCommon/BufferAllocator.h"
#include "ServerCommon/SimpleServerShutdownHandler.h"

#pragma hdrstop

#include <openssl/rand.h>
#include <openssl/err.h>
#include <openssl/ssl.h>

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CDebugTrace;
using JetByteTools::Core::CGlobalErrorHandler;

using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CAddressTypeIPv4;
using JetByteTools::Socket::CAddressType;
using JetByteTools::Socket::CConnectionLimiter;
using JetByteTools::Socket::ListenBacklog;

using JetByteTools::OpenSSL::CContext;

using JetByteTools::IO::CAsyncFileLog;

using std::cerr;
using std::endl;
using std::string;

typedef JetByteTools::Core::TZeroInitialiseExpandableBuffer<BYTE> ByteBuffer;

///////////////////////////////////////////////////////////////////////////////
// File level static variables
///////////////////////////////////////////////////////////////////////////////

// Note that this can be set from the command line using the -password command
// line argument. This should be the password that was used when you created
// your server's key file - see notes at the end of this file for how to
// create a self signed certificate and key.

static string serverKeyPassword = "password";

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static SSL_CTX *CreateSSLContext(
   const _tstring &certificatePath,
   bool verifyPeer);

///////////////////////////////////////////////////////////////////////////////
// Program entry point
///////////////////////////////////////////////////////////////////////////////

int main(int argc, const char *argv[ ])
{
   CGlobalErrorHandler errorHandler;

   auto result = 0;

   try
   {
      CSimpleServerShutdownHandler shutdownHandler;

      CAsyncFileLog log(_T("SecureWebSocketServer.log"));

      CDebugTrace::LogInstaller logInstaller(log);

      try
      {
         CWebSocketServerCommandLine commandLine(argc, argv);

         if (commandLine.Parse())
         {
            serverKeyPassword = CStringConverter::TtoA(commandLine.Password());

            CContext sslContext(CreateSSLContext(commandLine.CertificatePath(), commandLine.RequireClientCerts()));

            CIOPool pool(
               commandLine.NumberOfIOThreads(),
               commandLine.DisplayDebug());

            pool.Start();

            CSequencedStreamSocketAllocator socketAllocator(
               commandLine.SocketPoolSize(),
               commandLine.MaxBuffersPooledPerSocket(),
               commandLine.MaxConsecutiveOperationsOnOneSocket(),
               commandLine.DisplayDebug());

            CBufferAllocator bufferAllocator(
               commandLine.BufferSize(),
               commandLine.BufferPoolSize(),
               commandLine.DisplayDebug());

            const auto addressPreference = CAddressTypeIPv4::instance;

            const CFullAddress address(
               commandLine.Server(),
               commandLine.Port(),
               addressPreference);

            const auto listenBacklog = commandLine.ListenBacklog();

            CConnectionLimiter connectionLimiter(commandLine.MaxConnections());

            CSocketServer server(
               sslContext,
               commandLine.RequireClientCerts(),
               address,
               listenBacklog,
               pool,
               socketAllocator,
               bufferAllocator,
               connectionLimiter);

            server.Start();

            server.StartAcceptingConnections();

            shutdownHandler.SetServerControl(server, commandLine.EventNamePrefix());

            shutdownHandler.WaitForShutdownRequest();

            server.WaitForShutdownToComplete();

            pool.WaitForShutdownToComplete();

            socketAllocator.ReleaseSockets();

            bufferAllocator.Flush();
         }
         else
         {
            result = 1;
         }
      }
      catch(const CException &e)
      {
         OutputEx(_T("Exception: ") + e.GetDetails());

         result = 1;
      }
      catch(...)
      {
         OutputEx(_T("Unexpected exception"));

         result = 1;
      }
   }
   catch(const CException &e)
   {
      cerr << "Exception: " << e.GetDetailsA() << endl;

      result = 1;
   }
   catch(...)
   {
      cerr << "Unexpected exception" << endl;

      result = 1;
   }

   return result;
}

///////////////////////////////////////////////////////////////////////////////
// OpenSSL specific stuff.
//
// Note: The following code just sets up an example SSL Context, it's been
// put together from code samples in the OpenSSL distribution. You should read
// and understand the OpenSSL documentation and then write some REAL code to
// initialise your context in the way that you require.
///////////////////////////////////////////////////////////////////////////////

// This callback function is used to return the password for the server's key.
// In this example we simply return the value that was passed on the command line
// with the -password command line arg, or "password" if none is specified.

static int password_cb(
   char *buf,
   int num,
   int /*rwflag*/,
   void * /*userdata*/)
{
   const auto length = static_cast<int>(serverKeyPassword.length());

   if (num < (length + 1))
   {
      return 0;
   }

   strcpy_s(buf, num, serverKeyPassword.c_str());

   return length;
}

static SSL_CTX *CreateSSLContext(
   const _tstring &certificatePath,
   const bool verifyPeer)
{
   if (1 != RAND_status())
   {
      // this is now done automatically, see OpenSSL documentation if you ever get this message!

      throw JetByteTools::OpenSSL::CException(
         _T("CreateSSLContext() - random number generator not seeded"),
         0);
   }

   const auto pMethod = TLS_server_method();

   const auto pContext = SSL_CTX_new(pMethod);

   const string cipherList("ALL");

   if (0 == SSL_CTX_set_cipher_list(pContext, cipherList.c_str()))
   {
      const auto error = ERR_get_error();

      throw JetByteTools::OpenSSL::CException(
         _T("CreateSSLContext() - Failed to set cipher list to \"") +
         CStringConverter::AtoT(cipherList) + _T("\""), error);
   }

   const auto serverPemFile = CStringConverter::TtoA(certificatePath) + "\\server.pem";

   if (!(SSL_CTX_use_certificate_chain_file(pContext, serverPemFile.c_str())))
   {
      const auto error = ERR_get_error();

      throw JetByteTools::OpenSSL::CException(
         _T("CreateSSLContext() - Failed to load server certificate file: \"") +
         CStringConverter::AtoT(serverPemFile) + _T("\""), error);
   }

   SSL_CTX_set_default_passwd_cb(pContext, password_cb);

   if (!(SSL_CTX_use_PrivateKey_file(pContext, serverPemFile.c_str(), SSL_FILETYPE_PEM)))
   {
      const auto error = ERR_get_error();

      // Most common cause of failure here is that the password doesn't match...

      throw JetByteTools::OpenSSL::CException(
         _T("CreateSSLContext() - Failed to load private key file: \"") +
         CStringConverter::AtoT(serverPemFile) + _T("\""), error);
   }

   if (verifyPeer)
   {
      // To be able to verify a peer certificate we need a set of trusted root certificates...

      const auto rootPemFile = CStringConverter::TtoA(certificatePath) + "\\root.pem";

      /* Load the CAs we trust*/
      if(!(SSL_CTX_load_verify_locations(pContext, rootPemFile.c_str(), nullptr)))
      {
         const auto error = ERR_get_error();

         throw JetByteTools::OpenSSL::CException(
            _T("CreateSSLContext() - Failed to load root CA file: \"") +
            CStringConverter::AtoT(rootPemFile) + _T("\""), error);
      }

      if (!SSL_CTX_set_default_verify_paths(pContext))
      {
         const auto error = ERR_get_error();

         throw JetByteTools::OpenSSL::CException(
            _T("CreateSSLContext() - Failed to set verification paths"), error);
      }
   }

   return pContext;
}

// Making test certificates

// The following generates a single self signed certificate - which is enough to get
// going with. You may want something more complex, in which case read and understand
// the OpenSSL docs, the "Network Security with OpenSSL" book or various online
// resources.

// Make a key
// openssl genrsa -des3 -out server-k.pem 2048
// Note that the password used for the key should be passed to the server on the command line.
// If you leave out the -des3 and you will NOT protect the key with a password, the password
// is pretty pointless if you have to either code it into the server or pass it to the server
// on the command line...
//
// Make a cert
// openssl.exe req -new -x509 -key server-k.pem -out server-c.pem -days 10000 -config openssl.cnf
//
// Make a combined pem file for loading by the CreateSSLContext() function above.
// type server-c.pem server-k.pem > server.pem

// Creating a test cert for TLS 1.2 with ECDHE-ECDSA-AES128-GCM-SHA256

// openssl ecparam -name secp384r1 -genkey -out server-k.pem
// openssl req -new -x509 -key server-k.pem -out server-c.pem -days 730 -config openssl.cnf
// type private-k.pem server-c.pem > server.pem

// Convert to pfx for installation in windows cert store

// openssl pkcs12 -export -out server.pfx -inkey server-k.pem -in server-c.pem


///////////////////////////////////////////////////////////////////////////////
// End of file: ServerMain.cpp
///////////////////////////////////////////////////////////////////////////////
