///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#include "JetByteTools/WebSocketTools/AutoDetectProtocolHandler.h"
#include "JetByteTools/WebSocketTools/Headers.h"
#include "JetByteTools/WebSocketTools/Hixie76/ProtocolHandler.h"
#include "JetByteTools/WebSocketTools/Hixie76/RequestHeaders.h"
#include "JetByteTools/WebSocketTools/HyBi/ProtocolHandler.h"
#include "JetByteTools/WebSocketTools/HyBi/CloseStatusCodes.h"
#include "JetByteTools/WebSocketTools/HyBi/RequestHeaders.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::IIndexedOpaqueUserData;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;

using JetByteTools::OpenSSL::CContext;
using JetByteTools::OpenSSL::CStreamSocketServerEx;
using JetByteTools::Socket::ConnectionClosureReason;

using JetByteTools::WebSocket::CHeaders;
using JetByteTools::WebSocket::MessageStatus;
using JetByteTools::WebSocket::MessageType;
using JetByteTools::WebSocket::MessageTypeText;
using JetByteTools::WebSocket::ConnectionEstablishmentResult;
using JetByteTools::WebSocket::ConnectionEstablished;
using JetByteTools::WebSocket::IProtocolHandler;
using JetByteTools::WebSocket::MessageStatus;
using JetByteTools::WebSocket::MessageStatusComplete;
using JetByteTools::WebSocket::CAutoDetectProtocolHandler;
using JetByteTools::WebSocket::CAutoDetectProtocolHandlerAllocator;
using JetByteTools::WebSocket::IWebSocket;

using JetByteTools::WebSocket::HyBi::CloseStatusNormal;

typedef JetByteTools::WebSocket::Hixie76::CProtocolHandler CHixie76ProtocolHandler;
typedef JetByteTools::WebSocket::Hixie76::CRequestHeaders CHixie76RequestHeaders;
typedef JetByteTools::WebSocket::HyBi::CProtocolHandler CHyBiProtocolHandler;
typedef JetByteTools::WebSocket::HyBi::CRequestHeaders CHyBiRequestHeaders;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   const CContext &sslContext,
   const bool verifyPeer,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(sslContext, verifyPeer, address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_protocolHandlerIndex(socketAllocator.RequestUserDataSlot(_T("m_protocolHandlerIndex"))),
      m_bufferAllocator(bufferAllocator),
      m_protocolHandlerAllocator(*static_cast<JetByteTools::WebSocket::HyBi::IWebSocketServer *>(this), bufferAllocator)
{
   m_protocolHandlerAllocator.SupportHixie76(
      *this,
      CHixie76ProtocolHandler::SecureConnection |
      CHixie76ProtocolHandler::AccumulateCompleteMessage);

   m_protocolHandlerAllocator.SupportHyBiVersions(
      CAutoDetectProtocolHandlerAllocator::HyBiProtocolAllVersions,
      *this, 
      CHyBiProtocolHandler::SecureConnection | 
      CHyBiProtocolHandler::AccumulateCompleteMessage);
}

CSocketServer::~CSocketServer()
{
   WaitForShutdownToComplete();
}

void CSocketServer::OnSecureConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   Output(_T("OnSecureConnectionEstablished"));

   auto pHandler = m_protocolHandlerAllocator.Allocate(socket, true);

   socket.SetUserPointer(m_protocolHandlerIndex, pHandler);

   pHandler->AcceptHandshake();
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      //DEBUG_ONLY(Output(_T("WebSocket Data: \r\n") + DumpData(buffer.GetMemory(), buffer.GetUsed(), 60, true)));

      auto pHandler = reinterpret_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

      pHandler->OnDataReceived(buffer);
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CSocketServer::OnConnectionReset(
   IStreamSocket &socket,
   const DWORD /*lastError*/)
{
   OutputEx(_T("OnSocketReleased"));

   auto pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   pHandler->OnConnectionClosed();
}

void CSocketServer::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason /*reason*/)
{
   auto pHandler = reinterpret_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   pHandler->OnConnectionClosed();
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   Output(_T("OnSocketReleased"));

   auto pHandler = reinterpret_cast<IProtocolHandler *>(userData.GetUserPointer(m_protocolHandlerIndex));

   delete pHandler;
}

ConnectionEstablishmentResult CSocketServer::OnConnectionHandshake(
   IIndexedOpaqueUserData & /*userData*/,
   const string &uri,
   const bool secure,
   const CHeaders &requestHeaders,
   CHeaders &responseHeaders)
{
   Output(_T("OnConnectionHandshake - : ") + CStringConverter::AtoT(uri));

   CHixie76RequestHeaders::DumpRequestHeadersToDebug(requestHeaders);

   if (requestHeaders.Contains(CHixie76RequestHeaders::SecWebSocketKey1))
   {
      Output(_T("Protocol Version: Hixie76"));

      CHixie76RequestHeaders::AddSecWebSocketLocationHeader(responseHeaders, secure, requestHeaders, uri);
   }
   else
   {
      if (requestHeaders.Contains(CHyBiRequestHeaders::SecWebSocketVersion))
      {
         const auto version = requestHeaders.GetValue(CHyBiRequestHeaders::SecWebSocketVersion);

         Output(_T("Protocol Version: HyBi: ") + CStringConverter::AtoT(version));
      }
      else
      {
         throw CException(_T("CSocketServer::OnConnectionHandshake()"), _T("Protocol version unknown"));
      }
   }

   CHyBiRequestHeaders::AddOriginHeaders(responseHeaders, requestHeaders);

   return ::ConnectionEstablished;
}

void CSocketServer::OnConnectionEstablished(
   IWebSocket &socket,
   const string & /*uri*/,
   const bool /*secure*/)
{
   socket.TryRead();
}

void CSocketServer::OnMessage(
   IWebSocket &socket,
   const _tstring &text)
{
   // For this to be called you need to specify CProtocolHandler::DispatchTextAsStrings in the flags that are passed
   // during construction

   DEBUG_ONLY(Output(_T("OnMessage - \"") + text+ _T("\"")));

   socket.TryWriteText(text);

   socket.TryRead();
}

void CSocketServer::OnMessage(
   IWebSocket &socket,
   CSmartBuffer &buffer)
{
   // For this to be called you need to NOT specify CProtocolHandler::DispatchTextAsStrings in the flags that are passed
   // during construction

   DEBUG_ONLY(Output(_T("OnMessage - ") + ToString(buffer->GetUsed()) + _T(" bytes\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed())));

   // if we wanted to convert the buffer to text we could do this

   //const _tstring text = CStringConverter::UTF8toT(buffer->GetMemory(), buffer->GetUsed());

   //DEBUG_ONLY(Output(_T("OnMessage - \"") + text+ _T("\"")));

   socket.TryWriteText(buffer);

   socket.TryRead();
}

void CSocketServer::OnOutboundConnectionEstablished(
   JetByteTools::WebSocket::Hixie76::IWebSocket & /*socket*/,
   const CHeaders & /*responseHeaders*/)
{
   throw CException(
      _T("CSocketServer::OnOutboundConnectionEstablished()"),
      _T("Unexpected: we don't create outbound connections"));
}

void CSocketServer::OnConnectionEstablished(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket,
   const string &uri,
   const bool secure)
{
   Output(_T("OnConnectionEstablished - Hixie76: ") + CStringConverter::AtoT(uri));

   OnConnectionEstablished(static_cast<IWebSocket &>(socket), uri, secure);
}

void CSocketServer::OnData(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket,
   CSmartBuffer &buffer,
   const MessageStatus status)
{
   if (status != MessageStatusComplete)
   {
      throw CException(
         _T("CSocketServer::OnData()"),
         _T("Unexpected: message larger than buffer size"));
   }

   OnMessage(socket, buffer);
}

void CSocketServer::OnData(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket,
   const _tstring &text,
   const MessageStatus status)
{
   if (status != MessageStatusComplete)
   {
      throw CException(
         _T("CSocketServer::OnData()"),
         _T("Unexpected: message larger than buffer size"));
   }

   OnMessage(socket, text);
}

void CSocketServer::OnClientClose(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket)
{
   Output(_T("OnClientClose - Hixie76"));

   socket.Close();
}

void CSocketServer::OnConnectionEstablished(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   const string &uri,
   const bool secure)
{
   Output(_T("OnConnectionEstablished - HyBi: ") + CStringConverter::AtoT(uri));

   OnConnectionEstablished(static_cast<IWebSocket &>(socket), uri, secure);
}

void CSocketServer::OnOutboundConnectionEstablished(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const CHeaders & /*responseHeaders*/)
{
   throw CException(
      _T("CSocketServer::OnOutboundConnectionEstablished()"),
      _T("Unexpected: we don't create outbound connections"));
}

void CSocketServer::OnData(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   const _tstring &text,
   const MessageStatus status,
   const __int64 messageBytesOutstanding)
{
   if (status != MessageStatusComplete)
   {
      throw CException(
         _T("CSocketServer::OnData()"),
         _T("Unexpected: message larger than buffer size"));
   }

   if (messageBytesOutstanding != 0)
   {
      throw CException(
         _T("CSocketServer::OnData()"),
         _T("Unexpected: message larger than buffer size: ") + ToString(messageBytesOutstanding) + _T(" outstanding"));
   }


   OnMessage(socket, text);
}

void CSocketServer::OnData(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   CSmartBuffer &buffer,
   const MessageType type,
   const MessageStatus status,
   const __int64 messageBytesOutstanding)
{
   if (type == MessageTypeText)
   {
      if (status != MessageStatusComplete)
      {
         throw CException(
            _T("CSocketServer::OnData()"),
            _T("Unexpected: message larger than buffer size"));
      }

      if (messageBytesOutstanding != 0)
      {
         throw CException(
            _T("CSocketServer::OnData()"),
            _T("Unexpected: message larger than buffer size: ") + ToString(messageBytesOutstanding) + _T(" outstanding"));
      }

      OnMessage(socket, buffer);
   }
   else
   {
      throw CException(
         _T("CSocketServer::OnData()"),
         _T("Unexpected: we don't handle binary frames"));
   }
}

void CSocketServer::OnPingResponse(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const BYTE * /*pData*/,
   const BYTE /*length*/)
{
   throw CException(
      _T("CSocketServer::OnPingResponse()"),
      _T("Unexpected: we don't send pings"));
}

void CSocketServer::OnClientClose(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   const WORD status,
   const _tstring &reason)
{
   Output(_T("OnClientClose: HyBi:") + ToString(status) + _T(": ") + reason);

   socket.Close(CloseStatusNormal);
}

void CSocketServer::OnError(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const _tstring &message)
{
   OutputEx(_T("OnError:") + message);
}

void CSocketServer::OnClosed(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const WORD status)
{
   Output(_T("OnClosed:") + ToString(status));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
