///////////////////////////////////////////////////////////////////////////////
// File: HixieSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "HixieSocketServer.h"

#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/IOTools/Buffer.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#include "JetByteTools/WebSocketTools/AutoDetectProtocolHandler.h"
#include "JetByteTools/WebSocketTools/Headers.h"
#include "JetByteTools/WebSocketTools/Hixie76/ProtocolHandler.h"
#include "JetByteTools/WebSocketTools/Hixie76/RequestHeaders.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::IIndexedOpaqueUserData;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ConnectionClosureReason;

using JetByteTools::WebSocket::CHeaders;
using JetByteTools::WebSocket::MessageStatus;
using JetByteTools::WebSocket::ConnectionEstablishmentResult;
using JetByteTools::WebSocket::ConnectionEstablished;
using JetByteTools::WebSocket::IProtocolHandler;
using JetByteTools::WebSocket::MessageStatus;
using JetByteTools::WebSocket::MessageStatusComplete;

using JetByteTools::WebSocket::Hixie76::IWebSocket;
using JetByteTools::WebSocket::Hixie76::CProtocolHandler;
using JetByteTools::WebSocket::Hixie76::CRequestHeaders;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// CHixieSocketServer
///////////////////////////////////////////////////////////////////////////////

CHixieSocketServer::CHixieSocketServer(
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_protocolHandlerIndex(socketAllocator.RequestUserDataSlot(_T("m_protocolHandlerIndex"))),
      m_bufferAllocator(bufferAllocator),
      m_protocolHandlerAllocator(bufferAllocator)
{

}

CHixieSocketServer::~CHixieSocketServer()
{
   WaitForShutdownToComplete();
}

void CHixieSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   Output(_T("OnConnectionEstablished"));

   CProtocolHandler *pHandler = m_protocolHandlerAllocator.Allocate(*this, socket, CProtocolHandler::AccumulateCompleteMessage);

   socket.SetUserPointer(m_protocolHandlerIndex, pHandler);

   pHandler->AcceptHandshake();
}

void CHixieSocketServer::OnOutboundConnectionEstablished(
   IWebSocket & /*socket*/,
   const CHeaders & /*responseHeaders*/)
{
   throw CException(
      _T("CHixieSocketServer::OnOutboundConnectionEstablished()"),
      _T("Unexpected: we don't create outbound connections"));
}

void CHixieSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      //DEBUG_ONLY(Output(_T("WebSocket Data: \r\n") + DumpData(buffer.GetMemory(), buffer.GetUsed(), 60, true)));

      IProtocolHandler *pHandler = reinterpret_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

      pHandler->OnDataReceived(buffer);
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CHixieSocketServer::OnConnectionReset(
   IStreamSocket &socket,
   const DWORD /*lastError*/)
{
   IProtocolHandler *pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   pHandler->OnConnectionClosed();
}

void CHixieSocketServer::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason /*reason*/)
{
   IProtocolHandler *pHandler = reinterpret_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   pHandler->OnConnectionClosed();
}

void CHixieSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   Output(_T("OnSocketReleased"));

   IProtocolHandler *pHandler = reinterpret_cast<IProtocolHandler *>(userData.GetUserPointer(m_protocolHandlerIndex));

   delete pHandler;
}

ConnectionEstablishmentResult CHixieSocketServer::OnConnectionHandshake(
   IIndexedOpaqueUserData & /*userData*/,
   const string &uri,
   const bool secure,
   const CHeaders &requestHeaders,
   CHeaders &responseHeaders)
{
   Output(_T("OnConnectionHandshake - : ") + CStringConverter::AtoT(uri));

   CRequestHeaders::DumpRequestHeadersToDebug(requestHeaders);

   Output(_T("Protocol Version: Hixie76"));

   CRequestHeaders::AddSecWebSocketLocationHeader(responseHeaders, secure, requestHeaders, uri);

   CRequestHeaders::AddOriginHeaders(responseHeaders, requestHeaders);

   return ::ConnectionEstablished;
}

void CHixieSocketServer::OnConnectionEstablished(
   IWebSocket &socket,
   const string &uri,
   const bool /*secure*/)
{
   Output(_T("OnConnectionEstablished - Hixie76: ") + CStringConverter::AtoT(uri));

   socket.TryRead();
}

void CHixieSocketServer::OnData(
   IWebSocket &socket,
   CSmartBuffer &buffer,
   const MessageStatus status)
{
   // For this to be called you need to specify CProtocolHandler::DispatchTextAsStrings in the flags that are passed
   // during construction

   if (status != MessageStatusComplete)
   {
      throw CException(
         _T("CHixieSocketServer::OnData()"),
         _T("Unexpected: message larger than buffer size"));
   }

   // For this to be called you need to NOT specify CProtocolHandler::DispatchTextAsStrings in the flags that are passed
   // during construction

   //DEBUG_ONLY(Output(_T("OnMessage - ") + ToString(buffer->GetUsed()) + _T(" bytes\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed())));

   // if we wanted to convert the buffer to text we could do this

   //const _tstring text = CStringConverter::UTF8toT(buffer.GetMemory(), buffer.GetUsed());

   //DEBUG_ONLY(Output(_T("OnMessage - \"") + text+ _T("\"")));

   socket.TryWriteText(buffer);

   socket.TryRead();
}

void CHixieSocketServer::OnData(
   IWebSocket &socket,
   const _tstring &text,
   const MessageStatus status)
{
   // For this to be called you need to specify CProtocolHandler::DispatchTextAsStrings in the flags that are passed
   // during construction

   if (status != MessageStatusComplete)
   {
      throw CException(
         _T("CHixieSocketServer::OnData()"),
         _T("Unexpected: message larger than buffer size"));
   }

   //DEBUG_ONLY(Output(_T("OnData - \"") + text+ _T("\"")));

   socket.TryWriteText(text);

   socket.TryRead();
}

void CHixieSocketServer::OnClientClose(
   IWebSocket &socket)
{
   Output(_T("OnClientClose"));

   socket.Close();
}

///////////////////////////////////////////////////////////////////////////////
// End of file: HixieSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
