///////////////////////////////////////////////////////////////////////////////
// File: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2011 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "StreamSocketConnectionManager.h"
#include "TestMessages.h"

#include "JetByteTools/IOTools/IAllocateBuffers.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/WaitableCounter.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#include "JetByteTools/WebSocketTools/Headers.h"
#include "JetByteTools/WebSocketTools/Hixie76/IWebSocket.h"
#include "JetByteTools/WebSocketTools/HyBi/IWebSocket.h"
#include "JetByteTools/WebSocketTools/HyBi/ProtocolHandler.h"
#include "JetByteTools/WebSocketTools/HyBi/CloseStatusCodes.h"

#pragma hdrstop

#include <iostream>

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::ToString;
using JetByteTools::Core::Output;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::CException;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::ToStringA;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::IQueueTimers;
using JetByteTools::Core::ErrorCodeToErrorMessage;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IIOPool;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IAllocateBuffers;

using JetByteTools::Socket::CFilteringStreamSocketConnectionManager;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ConnectionClosureReason;
using JetByteTools::Socket::ConnectionAbortedMask;
using JetByteTools::Socket::GetClosureReasonAsString;

using JetByteTools::WebSocket::CHeaders;
using JetByteTools::WebSocket::MessageStatus;
using JetByteTools::WebSocket::MessageType;
using JetByteTools::WebSocket::ConnectionEstablishmentResult;
using JetByteTools::WebSocket::IProtocolHandler;
using JetByteTools::WebSocket::IWebSocket;

using JetByteTools::WebSocket::HyBi::CloseStatusNormal;

using std::string;
using std::cout;
using std::endl;

///////////////////////////////////////////////////////////////////////////////
// CStreamSocketConnectionManager
///////////////////////////////////////////////////////////////////////////////

CStreamSocketConnectionManager::CStreamSocketConnectionManager(
   const bool useHixie,
   const Milliseconds readTimeout,
   IIOPool &pool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator)
   :  CFilteringStreamSocketConnectionManager(*this, pool, socketAllocator, bufferAllocator),
      m_bufferAllocator(bufferAllocator),
      m_protocolHandlerIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_protocolHandlerIndex"))),
      m_timerHandleIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_timerHandleIndex"))),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CStreamSocketConnectionManager - m_userDataIndex"))),
      m_timeoutFilter(*this, m_timerQueue),
      m_readTimeout(readTimeout),
      m_hixieProtocolHandlerAllocator(bufferAllocator),
      m_hyBiProtocolHandlerAllocator(bufferAllocator),
      m_useHixie(useHixie)
{

}

CStreamSocketConnectionManager::~CStreamSocketConnectionManager()
{
   WaitForShutdownToComplete();
}

void CStreamSocketConnectionManager::OnPreOutgoingConnect(
   IIndexedOpaqueUserData &userData,
   const IAddress & /*address*/,
   const void *pUserData)
{
   const auto pTestMessages = const_cast<CTestMessages *>(static_cast<const CTestMessages*>(pUserData));

   userData.SetUserPointer(m_userDataIndex, pTestMessages);
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   if (m_useHixie)
   {
      auto pHandler = m_hixieProtocolHandlerAllocator.Allocate(
         *this,
         socket);

      socket.SetUserPointer(m_protocolHandlerIndex, pHandler);

      const auto timerHandle = m_timerQueue.CreateTimer();

      socket.SetUserData(m_timerHandleIndex, timerHandle);

      if (m_readTimeout != 0)
      {
         m_timeoutFilter.SetReadTimeout(socket, m_readTimeout, *this, 1);
      }

      const string host = "www.example.com";
      const string url = "\\";
      const string origin = "www.example.com";

      pHandler->Connect(host, url, origin);
   }
   else
   {
      auto pHandler = m_hyBiProtocolHandlerAllocator.Allocate(
         *this,
         socket);

      socket.SetUserPointer(m_protocolHandlerIndex, pHandler);

      const auto timerHandle = m_timerQueue.CreateTimer();

      socket.SetUserData(m_timerHandleIndex, timerHandle);

      if (m_readTimeout != 0)
      {
         m_timeoutFilter.SetReadTimeout(socket, m_readTimeout, *this, 1);
      }

      const string host = "www.example.com";
      const string url = "/";

      pHandler->Connect(host, url);
   }
}

void CStreamSocketConnectionManager::OnOutgoingConnectionFailed(
   IStreamSocket &socket,
   const IAddress & /*address*/,
   const DWORD lastError)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_connectionErrors[lastError]++;
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   pTestMessages->ConnectionComplete(false);

   socket.SetUserPointer(m_userDataIndex, nullptr);
}

void CStreamSocketConnectionManager::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   const auto pHandler = static_cast<IProtocolHandler *>(userData.GetUserPointer(m_protocolHandlerIndex));

   delete pHandler;

   const auto timerHandle = userData.GetUserData(m_timerHandleIndex);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      if (m_timerQueue.DestroyTimer(timerHandle))
      {
         OnError(_T("CStreamSocketConnectionManager::OnSocketReleased() - timer was set when destroyed, socket ref count bug!"));
      }
   }
}

void CStreamSocketConnectionManager::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      auto pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

      pHandler->OnDataReceived(buffer);
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CStreamSocketConnectionManager::OnConnectionClientClose(
   IStreamSocket & /*socket*/)
{
   Output(_T("OnConnectionClientClose"));
}

void CStreamSocketConnectionManager::OnConnectionReset(
   IStreamSocket &socket,
   const DWORD lastError)
{
   {
      ILockableObject::Owner lock(m_lock);

      m_connectionResetErrors[lastError]++;
   }

   auto pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   pHandler->OnConnectionClosed();
}

void CStreamSocketConnectionManager::OnError(
   const _tstring &message)
{
   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason reason)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   socket.SetUserData(m_timerHandleIndex, IQueueTimers::InvalidHandleValue);

   if (timerHandle != IQueueTimers::InvalidHandleValue)
   {
      m_timerQueue.DestroyTimerWithRefCountedUserData(timerHandle, &socket);
   }

   auto pHandler = static_cast<IProtocolHandler *>(socket.GetUserPointer(m_protocolHandlerIndex));

   pHandler->OnConnectionClosed();

   Output(_T("OnConnectionClosure: ") + GetClosureReasonAsString(reason));

   if (reason & ConnectionAbortedMask)
   {
      OnError(GetClosureReasonAsString(reason));
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      pTestMessages->MessagesComplete(!(reason & ConnectionAbortedMask));
   }
}

void CStreamSocketConnectionManager::ProcessData(
   IWebSocket &socket,
   CSmartBuffer &buffer)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      try
      {
         auto messageComplete = false;

         if (pTestMessages->ValidateReceivedData(buffer->GetMemory(), buffer->GetUsed(), messageComplete))
         {
            if (messageComplete && pTestMessages->SendAfterRecv())
            {
               SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
            }
         }
         else
         {
            #if (JETBYTE_TRACE_WEBSOCKET_READ_TO_DEBUG_LOG == 1 || JETBYTE_TRACE_WEBSOCKET_WRITE_TO_DEBUG_LOG == 1 || JETBYTE_DUMP_WEBSOCKET_READ_TO_DEBUG_LOG == 1 || JETBYTE_DUMP_WEBSOCKET_WRITE_TO_DEBUG_LOG == 1)
            OutputEx(_T("Reads complete: ") + socket.GetIdentifier());
            #endif

            const auto holdConnections = pTestMessages->HoldConnections();

            if (!holdConnections)
            {
               socket.Close();
            }
            else
            {
               pTestMessages->MessagesComplete(true);
            }
         }
      }
      catch (CException &e)
      {
         ILockableObject::Owner lock(m_lock);

         m_errors[e.GetWhat()]++;

         pTestMessages->MessagesComplete(false);

         throw;
      }
   }
}

void CStreamSocketConnectionManager::OnTimer(
   IQueueTimers::Timer::UserData userData)
{
   auto &socket = *reinterpret_cast<IWebSocket *>(userData);

   #if (JETBYTE_TRACE_WEBSOCKET_READ_TO_DEBUG_LOG == 1 || JETBYTE_TRACE_WEBSOCKET_WRITE_TO_DEBUG_LOG == 1 || JETBYTE_DUMP_WEBSOCKET_READ_TO_DEBUG_LOG == 1 || JETBYTE_DUMP_WEBSOCKET_WRITE_TO_DEBUG_LOG == 1)
   OutputEx(_T("on timer: ") + socket.GetIdentifier());
   #endif

   SendData(socket);

   socket.Release();
}

void CStreamSocketConnectionManager::OnTimer(
   IStreamSocket &socket,
   TimerCallback::UserData /*userData*/)
{
   OnError(_T("Read timeout: ") + ToString(static_cast<JetByteTools::Socket::IStreamSocketEx *>(&socket)));

   socket.AbortConnection();
}

void CStreamSocketConnectionManager::SendData(
   IWebSocket &socket)
{
   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   if (pTestMessages)
   {
      if (pTestMessages->WriteSome())
      {
         if (!pTestMessages->SendAfterRecv())
         {
            SetSendTimer(socket, pTestMessages->GetTransferRateMillis());
         }
      }
      else if (!pTestMessages->HoldConnections())
      {
         #if (JETBYTE_TRACE_WEBSOCKET_READ_TO_DEBUG_LOG == 1 || JETBYTE_TRACE_WEBSOCKET_WRITE_TO_DEBUG_LOG == 1 || JETBYTE_DUMP_WEBSOCKET_READ_TO_DEBUG_LOG == 1 || JETBYTE_DUMP_WEBSOCKET_WRITE_TO_DEBUG_LOG == 1)
         OutputEx(_T("Writes complete: ") + socket.GetIdentifier());
         #endif
         // Websocket closes are for the whole connection. We may still have data that we want to
         // receive and so we can't send a close here, we have to wait until we have got all the
         // data that we expect to get...
//         socket.Close();
      }
   }
}

void CStreamSocketConnectionManager::SetSendTimer(
   IWebSocket &socket,
   const Milliseconds timeout)
{
   const auto timerHandle = socket.GetUserData(m_timerHandleIndex);

   m_timerQueue.SetTimerWithRefCountedUserData(timerHandle, *this, timeout, &socket);
}

bool CStreamSocketConnectionManager::DumpErrors() const
{
   auto ok = true;

   if (!DumpErrors(_T("Connection establishment errors"), m_connectionErrors))
   {
      ok = false;
   }

   if (!DumpErrors(_T("Connections reset"), m_connectionResetErrors))
   {
      ok = false;
   }

   if (!m_errors.empty())
   {
      ok = false;

      cout << "Other errors" << endl;

      for (const auto &error : m_errors)
      {
         cout << ToStringA(error.second) << " - " << CStringConverter::TtoA(error.first) << endl;
      }
   }

   return ok;
}

bool CStreamSocketConnectionManager::DumpErrors(
   const _tstring &errorType,
   const LastErrors &lastErrors) const
{
   auto ok = true;

   if (!lastErrors.empty())
   {
      ok = false;

      cout << CStringConverter::TtoA(errorType) << endl;

      for (const auto &lastError : lastErrors)
      {
         cout << ToStringA(lastError.second) << " - " << ErrorCodeToErrorMessageA(lastError.first) << endl;
      }
   }

   return ok;
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   JetByteTools::WebSocket::Hixie76::IWebSocket & /*socket*/,
   const string & /*uri*/,
   const bool /*secure*/)
{
   // TODO - refactor the protocol handler so that we can use IWebSocketConnectionManager

   throw CException(_T("CStreamSocketConnectionManager::OnConnectionEstablished()"), _T("Unexpected. We only support outbound connections"));
}

void CStreamSocketConnectionManager::OnOutboundConnectionEstablished(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket,
   const CHeaders &responseHeaders)
{
   Output(_T("OnOutboundConnectionEstablished - Hixie76"));

   size_t index = 0;

   string header;

   Output(_T("Response headers"));

   while (responseHeaders.GetRawHeader(index, header))
   {
      Output(CStringConverter::AtoT(header));
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   pTestMessages->ConnectionComplete(true);

   pTestMessages->Connect(socket);

   SendData(socket);

   socket.TryRead();
}

void CStreamSocketConnectionManager::OnData(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket,
   CSmartBuffer &buffer,
   const MessageStatus /*status*/)
{
   ProcessData(socket, buffer);

   socket.TryRead();
}

void CStreamSocketConnectionManager::OnData(
   JetByteTools::WebSocket::Hixie76::IWebSocket & /*socket*/,
   const _tstring & /*text*/,
   const MessageStatus /*status*/)
{
   throw CException(_T("CStreamSocketConnectionManager::OnData()"), _T("Unexpected. We only dispatching text frames as bytes"));
}

void CStreamSocketConnectionManager::OnClientClose(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket)
{
   OutputEx(_T("OnClientClose"));

   socket.Close();
}

void CStreamSocketConnectionManager::OnOutboundConnectionEstablished(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   const CHeaders &responseHeaders)
{
   Output(_T("OnOutboundConnectionEstablished - HyBi"));

   size_t index = 0;

   string header;

   Output(_T("Response headers"));

   while (responseHeaders.GetRawHeader(index, header))
   {
      Output(CStringConverter::AtoT(header));
   }

   const string extensionsHeader = "sec-websocket-extensions";

   if (responseHeaders.Contains(extensionsHeader))
   {
      const auto extensions = responseHeaders.GetValue(extensionsHeader);

      if (string::npos != extensions.find("deflate-stream"))
      {
         Output(_T("Deflate stream enabled"));
      }
   }

   auto pTestMessages = static_cast<CTestMessages*>(socket.GetUserPointer(m_userDataIndex));

   pTestMessages->ConnectionComplete(true);

   pTestMessages->Connect(socket);

   SendData(socket);

   socket.TryRead();
}

ConnectionEstablishmentResult CStreamSocketConnectionManager::OnConnectionHandshake(
   IIndexedOpaqueUserData & /*userData*/,
   const string & /*uri*/,
   const bool /*secure*/,
   const CHeaders & /*requestHeaders*/,
   CHeaders & /*responseHeaders*/)
{
   // TODO - refactor the protocol handler so that we can use IWebSocketConnectionManager

   throw CException(_T("CStreamSocketConnectionManager::OnConnectionHandshake()"), _T("Unexpected. We only support outbound connections"));
}

void CStreamSocketConnectionManager::OnConnectionEstablished(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const string & /*uri*/,
   const bool /*secure*/)
{
   // TODO - refactor the protocol handler so that we can use IWebSocketConnectionManager

   throw CException(_T("CStreamSocketConnectionManager::OnConnectionEstablished()"), _T("Unexpected. We only support outbound connections"));
}

void CStreamSocketConnectionManager::OnData(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const _tstring & /*text*/,
   const MessageStatus /*status*/,
   const __int64 /*messageBytesOutstanding*/)
{
   throw CException(_T("CStreamSocketConnectionManager::OnData()"), _T("Unexpected. We only dispatching text frames as bytes"));
}

void CStreamSocketConnectionManager::OnData(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   CSmartBuffer &buffer,
   const MessageType /*type*/,
   const MessageStatus /*status*/,
   const __int64 /*messageBytesOutstanding*/)
{
   //DEBUG_ONLY(Output(_T("Application Data: \r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true)));

   ProcessData(socket, buffer);

   socket.TryRead();
}

void CStreamSocketConnectionManager::OnPingResponse(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   const BYTE *pData,
   const BYTE length)
{
   (void)socket;
   (void)pData;
   (void)length;
}

void CStreamSocketConnectionManager::OnClientClose(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket,
   const WORD status,
   const _tstring &reason)
{
   Output(_T("OnClientClose: ") + ToString(status) + _T(": ") + reason);

   socket.Close(CloseStatusNormal);
}

void CStreamSocketConnectionManager::OnError(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const _tstring &message)
{
   Output(_T("OnError: ") + message);

   ILockableObject::Owner lock(m_lock);

   m_errors[message]++;
}

void CStreamSocketConnectionManager::OnClosed(
   JetByteTools::WebSocket::HyBi::IWebSocket & /*socket*/,
   const WORD status)
{
   Output(_T("OnClosed: ") + ToString(status));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: StreamSocketConnectionManager.cpp
///////////////////////////////////////////////////////////////////////////////
