///////////////////////////////////////////////////////////////////////////////
// File: TestMessage.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "TestMessage.h"

#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"

#pragma hdrstop

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::Output;
using JetByteTools::Core::DumpData;

using std::min;

///////////////////////////////////////////////////////////////////////////////
// CTestMessage
///////////////////////////////////////////////////////////////////////////////

CTestMessage::CTestMessage(
   const DataLength messageSize,
   const bool textMessage)
   :  m_messageSize(messageSize),
      m_pData(new BYTE[messageSize]),
      m_sendOffset(0),
      m_receiveOffset(0),
      m_sendIndex(0),
      m_recvIndex(0)
{
   static const BYTE index = 1;

   if (textMessage)
   {
      for (size_t i = 0; i < m_messageSize - 1; ++i)
      {
         m_pData[i] = static_cast<BYTE>((i % 94) + 32);
      }

      m_pData[m_messageSize - 1] = 0;
   }
   else
   {
      const auto netOrder = (int)htonl(messageSize - sizeof(int));

      memcpy(m_pData, &netOrder, sizeof(int));

      for (auto i = sizeof(int); i < m_messageSize; ++i)
      {
         if (i % 2)
         {
            m_pData[i] = static_cast<BYTE>(i);
         }
         else
         {
            m_pData[i] = static_cast<BYTE>(index);
         }
      }
   }
}

CTestMessage::~CTestMessage()
{
   delete [] m_pData;
}

// Implement ITestMessage

void CTestMessage::ExpectRepeats(
   const size_t numberOfRepeats)
{
   m_times.resize(numberOfRepeats);
}

_tstring CTestMessage::GetIdentifier() const
{
   return _T("Test Message");
}

const CTestMessage::Times &CTestMessage::GetTimes() const
{
   return m_times;
}

const BYTE *CTestMessage::GetBytes() const
{
   if (m_sendOffset == 0 && m_receiveOffset != 0)
   {
      const_cast<CTestMessage *>(this)->ResetMessage();
   }

   return m_pData + m_sendOffset;
}

CTestMessage::DataLength CTestMessage::GetSize() const
{
   return m_messageSize - m_sendOffset;
}

static _tstring GetDataMismatch(
   const BYTE * const pData,
   const BYTE * const pExpectedData,
   const CTestMessage::DataLength size,
   const CTestMessage::DataLength contextLength,
   const CTestMessage::DataLength maxMismatchLength)
{
   auto dataMatches = true;

   CTestMessage::DataLength firstMismatch = 0;

   for (CTestMessage::DataLength i = 0; i < size && dataMatches; ++i)
   {
      if (pData[i] != pExpectedData[i])
      {
         firstMismatch = i;
         dataMatches = false;
      }
   }

   if (!dataMatches)
   {
      auto lastMismatch = size;

      for (auto i = firstMismatch; i < size && !dataMatches; ++i)
      {
         if (pData[i] == pExpectedData[i])
         {
            lastMismatch = i;
            dataMatches = true;
         }
      }

      if (firstMismatch > contextLength)
      {
         firstMismatch -= contextLength;
      }
      else
      {
         firstMismatch = 0;
      }

      if (lastMismatch - firstMismatch > maxMismatchLength)
      {
         lastMismatch = firstMismatch + maxMismatchLength;
      }
      else
      {
         if (size - lastMismatch > contextLength)
         {
            lastMismatch += contextLength;
         }
         else
         {
            lastMismatch = size;
         }
      }

      const size_t dumpLength = lastMismatch - firstMismatch + 1;

      auto result = _T("Got: ") + DumpData(pData + firstMismatch, dumpLength);

      result += _T(" Expected: ") + DumpData(pExpectedData + firstMismatch, dumpLength);

      return result;
   }

   return _T("");
}

bool CTestMessage::ValidateResponse(
   const BYTE * const pData,
   const DataLength dataLength,
   DataLength &dataUsed) const
{
   dataUsed = min(m_messageSize - m_receiveOffset, dataLength);

   const auto messageComplete = (m_receiveOffset + dataUsed == m_messageSize);

   if (messageComplete)
   {
      m_times[m_recvIndex].recvTime = m_counter.GetElapsedTimeAsDWORD();
   }

   if (0 != memcmp(pData, m_pData + m_receiveOffset, dataUsed))
   {
      const DataLength contextLength = 10;
      const DataLength maxMismatchLength = 20;

      const auto dataMismatch = GetDataMismatch(pData, m_pData + m_receiveOffset, dataUsed, contextLength, maxMismatchLength);

      Output(dataMismatch);

      Output(_T("     Got: ") + DumpData(pData, dataUsed));
      Output(_T("Expected: ") + DumpData(m_pData + m_receiveOffset, dataUsed));

      throw CException(_T("CTestMessage::ValidateResponse()"), _T("Data does not validate: ") + dataMismatch);
   }

   m_receiveOffset += dataUsed;

   return messageComplete;
}

CTestMessage::DataLength CTestMessage::GetRecvOffset() const
{
   return m_receiveOffset;
}

CTestMessage::DataLength CTestMessage::GetResponseSize() const
{
   return m_messageSize;
}

bool CTestMessage::BytesToSend(
   const DataLength bytesToSend) const
{
   auto allSent = false;

   if (m_sendOffset + bytesToSend == m_messageSize)
   {
      allSent = true;
   }

   return allSent;
}

bool CTestMessage::BytesSent(
   const DataLength bytesSent)
{
   auto allSent = false;

   m_sendOffset += bytesSent;

   if (m_sendOffset == m_messageSize)
   {
      m_sendOffset = 0;

      m_counter.Reset();

      m_times[m_sendIndex].sentTime = 0;
      m_times[m_sendIndex].recvTime = 0;

      m_sendIndex++;

      allSent = true;
   }
   else if (m_sendOffset > m_messageSize)
   {
      throw CException(_T("CTestMessage::BytesSent()"), _T("More bytes sent than there are in the message"));
   }

   return allSent;
}

void CTestMessage::ResetMessage()
{
   if (m_receiveOffset != m_messageSize)
   {
      throw CException(_T("CTestMessage::ResetMessage()"), _T("Message is not complete"));
   }

   m_receiveOffset = 0;

   m_recvIndex++;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: TestMessage.cpp
///////////////////////////////////////////////////////////////////////////////
