///////////////////////////////////////////////////////////////////////////////
// File: TestMessages.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2005 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "TestMessages.h"
#include "ITestMessage.h"
#include "MessageCollection.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"

#include "JetByteTools/WebSocketTools/Hixie76/IWebSocket.h"
#include "JetByteTools/WebSocketTools/HyBi/IWebSocket.h"

#pragma hdrstop

#include <iostream>

#include <algorithm>    // for min

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Milliseconds;

using JetByteTools::Core::CException;
using JetByteTools::Core::ToString;
using JetByteTools::Core::_tstring;
using JetByteTools::Core::ILockableObject;
using JetByteTools::Core::CStringConverter;

using std::min;

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static DWORD GetAverageRTTFromTimes(
   const ITestMessage::Times &times);

static DWORD GetMaxRTTFromTimes(
   const ITestMessage::Times &times);

static DWORD GetMinRTTFromTimes(
   const ITestMessage::Times &times);

///////////////////////////////////////////////////////////////////////////////
// CTestMessages
///////////////////////////////////////////////////////////////////////////////

CTestMessages::CTestMessages(
   const size_t connectionIndex,
   const size_t repeatCount,
   const Milliseconds transferRateMillis,
   CMessageCollection &messageCollection,
   const bool sendAfterRecv,
   const bool holdConnections,
   const DataLength skipSignOnMessageBytes,
   const bool fragments,
   const bool multipleMessages,
   const TrackPending trackPending,
   const bool textMessages)
   :  m_messageCollection(messageCollection),
      m_connectionIndex(connectionIndex),
      m_sendAfterRecv(sendAfterRecv),
      m_holdConnections(holdConnections),
      m_skipSignOnMessageBytes(skipSignOnMessageBytes),
      m_fragments(fragments),
      m_multipleMessages(multipleMessages),
      m_sendIndex(0),
      m_receiveIndex(0),
      m_succeeded(false),
      m_connected(false),
      m_repeatCount(repeatCount),
      m_transferRateMillis(transferRateMillis),
      m_currentSendIteration(0),
      m_currentReceiveIteration(0),
      m_signOnBytesSkipped(0),
      m_complete(false),
      m_connectionComplete(false),
      m_allDataSent(false),
      m_dataPending(0),
      m_trackPending(trackPending),
      m_textMessages(textMessages),
      m_pHyBiSocket(nullptr),
      m_pSocket(nullptr)
{
}

void CTestMessages::AddMessage(
   ITestMessage *pMessage)
{
   if (!pMessage)
   {
      throw CException(_T("CTestMessages::AddMessage("), _T("pMessage must not be null"));
   }

   pMessage->ExpectRepeats(m_repeatCount);

   m_messages.push_back(pMessage);
}

CTestMessages::~CTestMessages()
{
   for (auto message : m_messages)
   {
      delete message;
   }
}

void CTestMessages::Connect(
   JetByteTools::WebSocket::Hixie76::IWebSocket &socket)
{
   if (m_pSocket)
   {
      throw CException(_T("CTestMessages::Connect()"), _T("Already connected"));
   }

   m_pSocket = &socket;
}

void CTestMessages::Connect(
   JetByteTools::WebSocket::HyBi::IWebSocket &socket)
{
   if (m_pSocket)
   {
      throw CException(_T("CTestMessages::Connect()"), _T("Already connected"));
   }

   m_pSocket = &socket;

   m_pHyBiSocket = &socket;
}

bool CTestMessages::ValidateReceivedData(
   const BYTE *pData,
   const DataLength dataLength,
   bool &messageComplete)
{
   ILockableObject::Owner lock(m_lock);

   if (m_signOnBytesSkipped == m_skipSignOnMessageBytes)
   {
      const auto numberOfDistinctMessages = m_messages.size();

      auto pMessageData = pData;

      auto dataAvailable = dataLength;

      if (m_trackPending == TrackPendingBytes)
      {
         m_dataPending -= dataAvailable;
      }

      while (dataAvailable != 0)
      {
         if (m_receiveIndex == numberOfDistinctMessages)
         {
            throw CException(_T("CTestMessages::ValidateReceivedData()"), _T("No more messages to validate"));
         }

         auto pMessage = m_messages[m_receiveIndex];

         DataLength dataUsed = 0;

         if (pMessage->ValidateResponse(pMessageData, dataAvailable, dataUsed))
         {
            m_receiveIndex++;

            if (m_trackPending == TrackPendingMessages)
            {
               m_dataPending--;
            }

            if (m_receiveIndex == numberOfDistinctMessages)
            {
               m_currentReceiveIteration++;

               if (m_repeatCount != m_currentReceiveIteration)
               {
                  pMessage->ResetMessage();

                  m_receiveIndex = 0;
               }
            }
         }

         dataAvailable -= dataUsed;

         pMessageData += dataUsed;
      }

      messageComplete = !m_dataPending;

      return numberOfDistinctMessages != m_receiveIndex;
   }
   else
   {
      const auto bytesToSkip = min(m_skipSignOnMessageBytes - m_signOnBytesSkipped, dataLength);

      m_signOnBytesSkipped += bytesToSkip;

      if (dataLength - bytesToSkip > 0)
      {
         return ValidateReceivedData(pData + bytesToSkip, dataLength - bytesToSkip, messageComplete);
      }
   }

   return true;
}

static int RandomValueInRange(
   const int lowerBound,
   const int upperBound)
{
   return static_cast<int>(static_cast<double>(rand()) / (RAND_MAX + 1) * (upperBound - lowerBound) + lowerBound);
}

int CTestMessages::GetNumberOfMessagesToSend(
   const size_t numMessages,
   const int minMessages) const
{
   const auto rangeMax = static_cast<int>(min(numMessages, static_cast<size_t>(std::numeric_limits<int>::max())) - m_sendIndex + (numMessages * (m_repeatCount - 1)));

   return RandomValueInRange(minMessages, rangeMax);
}

bool CTestMessages::WriteSome()
{
   ILockableObject::Owner lock(m_lock);

   if (!m_allDataSent)
   {
      const auto numberOfDistinctMessages = m_messages.size();

      const auto totalNumberOfMessagesToSend = numberOfDistinctMessages * m_repeatCount;

      const auto messagesAlreadySent = (numberOfDistinctMessages * m_currentSendIteration) + m_sendIndex;

      const auto numMessages = totalNumberOfMessagesToSend - messagesAlreadySent;

      if (m_repeatCount != 0)
      {
         const auto minMessages = (m_sendAfterRecv && m_trackPending == TrackPendingMessages) ? 2 : 1;

         auto messagesToSend = m_multipleMessages ? GetNumberOfMessagesToSend(numMessages, minMessages) : 1;

         while (messagesToSend != 0 && m_sendIndex != numberOfDistinctMessages)
         {
            const auto lastMessage = (m_repeatCount == m_currentSendIteration && m_sendIndex == numberOfDistinctMessages - 1);

            auto pMessage = m_messages[m_sendIndex];

            if (messagesToSend == 1 && m_fragments && !lastMessage)
            {
               // Send part of a message

               const DataLength bytesToSend = RandomValueInRange(1, pMessage->GetSize());

               if (m_trackPending == TrackPendingBytes)
               {
                  m_dataPending += bytesToSend;
               }

               const auto pBytes = pMessage->GetBytes();

               if (pMessage->BytesSent(bytesToSend))
               {
                  if (m_trackPending == TrackPendingMessages)
                  {
                     m_dataPending++;
                  }

                  m_sendIndex++;
               }

               // All manipulation of tracking data MUST occur before we call
               // TryWrite() as if this completes "inline" and a corresponding Recv
               // also completes "inline" then the validation of the recv's data will
               // occur on this thread before the TryWrite returns to us.

               if (m_textMessages)
               {
                  const auto pData = reinterpret_cast<const char *>(pBytes);

                  const auto skipTrailingNull = (pData[bytesToSend] == 0);

                  m_pSocket->TryWriteText(CStringConverter::AtoT(pData, bytesToSend - skipTrailingNull ? 1 : 0));
               }
               else
               {
                  if (m_pHyBiSocket)
                  {
                     m_pHyBiSocket->TryWriteBinary(pBytes, bytesToSend);
                  }
                  else
                  {
                     throw "TODO";
                  }
               }
            }
            else
            {
               // send a whole message...

               const auto bytesToSend = pMessage->GetSize();

               const auto pBytes = pMessage->GetBytes();

               if (m_trackPending == TrackPendingBytes)
               {
                  m_dataPending += bytesToSend;
               }

               if (pMessage->BytesSent(bytesToSend))
               {
                  if (m_trackPending == TrackPendingMessages)
                  {
                     m_dataPending++;
                  }

                  m_sendIndex++;
               }

               if (m_textMessages)
               {
                  m_pSocket->TryWriteText(CStringConverter::AtoT(reinterpret_cast<const char *>(pBytes), bytesToSend - 1));
               }
               else
               {
                  if (m_pHyBiSocket)
                  {
                     m_pHyBiSocket->TryWriteBinary(pBytes, bytesToSend);
                  }
                  else
                  {
                     throw "TODO";
                  }
               }
            }

            messagesToSend--;

            if (m_sendIndex == numberOfDistinctMessages)
            {
               m_currentSendIteration++;

               if (m_repeatCount != m_currentSendIteration)
               {
                  m_sendIndex = 0;
               }
            }
         }
      }

      if (m_repeatCount == m_currentSendIteration &&
          m_sendIndex == numberOfDistinctMessages)
      {
         m_allDataSent = true;
      }
   }

   return !m_allDataSent;
}

void CTestMessages::ConnectionComplete(
   const bool succeeded)
{
   if (m_connectionComplete.ToggleIfFalse())
   {
      InternalConnectionComplete(succeeded);
   }
}

void CTestMessages::InternalConnectionComplete(
   const bool succeeded)
{
   m_connected = succeeded;

   m_messageCollection.ConnectionComplete(succeeded);

   if (!succeeded)
   {
      MessagesComplete(false);
   }
}

void CTestMessages::MessagesComplete(
   const bool succeeded)
{
   if (m_connectionComplete.ToggleIfFalse())
   {
      InternalConnectionComplete(false);
   }
   else if (m_complete.ToggleIfFalse())
   {
      m_succeeded = succeeded;

      m_messageCollection.MessagesComplete();
   }
}

bool CTestMessages::Succeeded(
   StringErrors &errors) const
{
   if (m_connected)
   {
      const auto numberOfDistinctMessages = m_messages.size();

      const auto lastMessage = numberOfDistinctMessages - 1;

      const auto pMessage = m_messages[m_receiveIndex == numberOfDistinctMessages ? lastMessage : m_receiveIndex];

      const auto lastMessageSize = pMessage->GetResponseSize();

      const auto lastMessageData = pMessage->GetRecvOffset();

      if (m_succeeded)
      {
         if (m_receiveIndex != numberOfDistinctMessages ||
             lastMessageData != lastMessageSize)
         {
            const auto message = (m_receiveIndex == lastMessage ? _T("Incomplete final message: ") :
               _T("Incomplete response RecvIndex: ") + ToString(m_receiveIndex)) +
               _T(" Procesed: ") + ToString(lastMessageData) + _T(" bytes Expected: ") +
               ToString(lastMessageSize) + _T(" bytes");

            errors[message]++;
         }
      }
      else
      {
         const auto message = _T("Failed Messages - SendIndex: ") + ToString(m_sendIndex) +
            (m_repeatCount != 1 ? _T(" current send iteration: ") + ToString(m_currentSendIteration) +
            _T(" of ") + ToString(m_repeatCount) : _T(""));

         errors[message]++;

         if (m_receiveIndex != numberOfDistinctMessages)
         {
            const auto message = (m_receiveIndex == lastMessage ? _T("Incomplete final message: ") :
               _T("Incomplete response RecvIndex: ") + ToString(m_receiveIndex)) +
               _T(" Procesed: ") + ToString(lastMessageData) + _T(" bytes Expected: ") +
               ToString(lastMessageSize) + _T(" bytes");

            errors[message]++;
         }
      }
   }

   return m_succeeded;
}

Milliseconds CTestMessages::GetTransferRateMillis() const
{
   return m_transferRateMillis;
}

bool CTestMessages::SendAfterRecv() const
{
   return !m_allDataSent && m_sendAfterRecv;
}

bool CTestMessages::HoldConnections() const
{
   return m_holdConnections;
}

DWORD CTestMessages::GetAverageRTT() const
{
   DWORD averageRTT = 0;

   DWORD numMessages = 0;

   for (auto message : m_messages)
   {
      const auto &times = message->GetTimes();

      const auto rtt = GetAverageRTTFromTimes(times);

      if (rtt != 0)
      {
         averageRTT += rtt;

         numMessages++;
      }
   }

   return numMessages == 0 ? 0 : (averageRTT / numMessages);
}

DWORD CTestMessages::GetMaxRTT() const
{
   DWORD maxRTT = 0;

   for (auto message : m_messages)
   {
      const auto &times = message->GetTimes();

      const auto rtt = GetMaxRTTFromTimes(times);

      if (rtt > maxRTT)
      {
         maxRTT = rtt;
      }
   }

   return maxRTT;
}

DWORD CTestMessages::GetMinRTT() const
{
   DWORD minRTT = MAXDWORD;

   for (auto message : m_messages)
   {
      const auto &times = message->GetTimes();

      const auto rtt = GetMinRTTFromTimes(times);

      if (rtt < minRTT)
      {
         minRTT = rtt;
      }
   }

   return minRTT;
}

///////////////////////////////////////////////////////////////////////////////
// Static helper methods
///////////////////////////////////////////////////////////////////////////////

static DWORD GetAverageRTTFromTimes(
   const ITestMessage::Times &times)
{
   size_t averageRTT = 0;

   size_t numMessages = 0;

   for (auto time : times)
   {
      const auto rtt = time.GetRoundTripTime();

      if (rtt != 0)
      {
         averageRTT += rtt;
         numMessages++;
      }
   }

   return numMessages == 0 ? 0 : static_cast<DWORD>(averageRTT / numMessages);
}

static DWORD GetMaxRTTFromTimes(
   const ITestMessage::Times &times)
{
   DWORD maxRTT = 0;

   for (auto time : times)
   {
      const auto rtt = time.GetRoundTripTime();

      if (rtt > maxRTT)
      {
         maxRTT = rtt;
      }
   }

   return maxRTT;
}

static DWORD GetMinRTTFromTimes(
   const ITestMessage::Times &times)
{
   DWORD minRTT = MAXDWORD;

   for (auto time : times)
   {
      const auto rtt = time.GetRoundTripTime();

      if (rtt < minRTT)
      {
         minRTT = rtt;
      }
   }

   return minRTT;
}

///////////////////////////////////////////////////////////////////////////////
// End of file: TestMessages.cpp
///////////////////////////////////////////////////////////////////////////////
